import types, sys, os, os.path, stat, string, fnmatch




from Directories import Set as DS
from Keywords import Keyword as Kw




verbose			= False
omitExecPath	= True  # Omit (whenever possible) the path to the executable if the latter is in the PATH
skipMissExec	= True  # Skip the entries whose executables are not found




Console			= Kw("Console")
XWindow			= Kw("XWindow")
GNOME			= Kw("GNOME")
KDE				= Kw("KDE")
XFce			= Kw("XFce")




skip = {Console : False, XWindow : False, GNOME : False, KDE : False, XFce : False}




paths = DS(*os.environ["PATH"].split(":"))




class NotFound(Exception) :
	"""Raised when an appropriate file is not found"""




def msg(s, newline = True) :
	if verbose :
		sys.stderr.write(s)
		if newline :
			sys.stderr.write("\n")




def warn(s, newline = True) :
	sys.stderr.write(s)
	if newline :
		sys.stderr.write("\n")




def fatal(s, newline = True, code = 1) :
	sys.stderr.write(s)
	if newline :
		sys.stderr.write("\n")
	sys.exit(code)




def isExec(name) :
	"""Check whether the specified name can be considered executable"""
	# Exec's have x-bit set for current uid, aren't directories, and don't have leading dot
	try :
		return (stat.S_IMODE(os.stat(name)[stat.ST_MODE]) & 0100) \
			and not os.path.isdir(name) \
			and not os.path.basename(name).startswith(".")
	except OSError :
		return False




def locateExec(name) :
	"""Locate the executable in paths specified by PATH variable.
		If full path is specified, then check it first and if this fails, scan over PATH.
		Return full path to the executable or raise NotFound"""
	if os.path.isabs(name) :
		if isExec(name) :
			return name
	for p in paths :
		x = os.path.join(p, name)
		if isExec(x) :
			return x
	raise NotFound(name)




class Prefixes(DS) :
	"""Class to manipulate the paths' prefixes"""

	binDirs = ("bin", "sbin", "games")

	def __add__(self, other) :
		return Prefixes( *(tuple(self) + tuple(other)) )

	def valid(self, prefix) :
		"""Check whether the specified prefix can be considered valid"""
		bins = self.binDirs + ("",)
		if os.path.isdir(prefix) :
			for d in bins :
				dir = os.path.join(prefix, d)
				if os.path.isdir(dir) :
					for f in os.listdir(dir) :
						if isExec(os.path.join(dir, f)) :
							return True
		return False




def _fromEnv() :
	"""Return Prefixes instance with the valid prefixes extracted from the environment"""
	prefixes = []
	for x in os.environ["PATH"].split(":") :
		# FIXME : detect and kill the trailing slash in path
		b = os.path.basename(x)
		d = os.path.dirname(x)
		if b in Prefixes.binDirs :
			x = d
		if not d in prefixes :
			prefixes.append(x)
	return Prefixes(*prefixes)




prefixes = _fromEnv() + Prefixes("/", "/usr", "/usr/local", "/usr/X11", "/usr/X11R6", "/opt", "~")




class Application(object) :
	"""Single application that has specific executable and thus can be put into a menu"""
	# Native environment for the application
	Console	= 0
	XWindow	= 1
	GNOME	= 2
	KDE		= 3
	XFce	= 4
	# GNUstep???

	def __init__(self) :
		# Attributes to set, mandatory(+), optional(-) :
		# +exename		-- executable, with or w/o path
		# -exeargs		-- arguments that follow the exename
		# +execmd		-- a fully formatted command line for the app
		# +keywords		-- a set of keywords, must not be empty
		# +nativenv		-- the environment required by the app
		# -name			-- short description of the app
		# -comment		-- a more detailed description

		# NOTE : When determining whether the application should be run in terminal,
		# consult the self.terminal and _not_ the self.nativenv

		# If an (optional) name attr is missing, deduce it from the executable
		try :
			self.name
		except AttributeError :
			self.name = os.path.basename(self.exename).capitalize()

		# If an absolute path is given, check it
		# otherwise try to locate the executable via the PATH envar
		path = self.exename
		if os.path.isabs(path) :
			if skipMissExec and not isExec(path) :
				raise NotFound(path + " not found")
			if omitExecPath :
				try :
					# Check whether the specified filename is that one accessible via the PATH
					located = locateExec(path)
					try :
						if os.path.samefile(path, located) :
							path = os.path.basename(path)
					except OSError :
						pass
				except NotFound :
					pass
		else :
			# Here and forth we're given a relative path
			try :
				located = locateExec(path)
				# There is no reason to check for located<->path identity; we simply
				# make sure the executable can be accessed via PATH
				if not omitExecPath :
					# Sice we're initially given a relative path, make it full
					# by consuming the result of locateExec()
					path = located
			except NotFound :
				if skipMissExec :
					raise NotFound(path + " not found in PATH")

		# Execmd is fully formatted command line for the application
		try :
			self.execmd = path + " " + self.exeargs # exeargs may be missing, it's OK
		except AttributeError :
			self.execmd = path

		if self.terminal :
			self.nativenv = Application.Console
		else :
			self.nativenv = Application.XWindow

		for x in self.keywords :
			if   fnmatch.fnmatchcase(x, "*GNOME*") :
				self.nativenv = Application.GNOME
				break
			elif fnmatch.fnmatchcase(x, "*KDE*") :
				self.nativenv = Application.KDE
				break
			elif fnmatch.fnmatchcase(x, "*XFCE*") :
				self.nativenv = Application.XFce
				break
			# TODO : GNUstep???

		# Not all XFce apps have a distinguishing keywords therefore make an
		# additional check for the clues from its public name
		if self.nativenv == Application.XWindow and fnmatch.fnmatch(self.name.lower(), "*xfce*") :
			self.nativenv = Application.XFce

		# Skip unwanted entries
		if self.nativenv == Application.Console and skip[Console] :
			raise NotFound("don't want console app")
		if self.nativenv == Application.XWindow and skip[XWindow] :
			raise NotFound("don't want X11 app")
		if self.nativenv == Application.GNOME and skip[GNOME] :
			raise NotFound("don't want GNOME app")
		if self.nativenv == Application.KDE and skip[KDE] :
			raise NotFound("don't want KDE app")
		if self.nativenv == Application.XFce and skip[XFce] :
			raise NotFound("don't want XFce app")

		self.hash = hash(self.exename) # FIXME : is this enough?

	def __hash__(self) :
		return self.hash




def merge(entries) :
	"""Removes repeating entries from the list using fuzzy matching technique"""
	# NOTE : this is the deliberate violation of the immutability principle
	# __matched attribute is set for the entries that have been successfully matched before
	# Such entries should be skipped
	# It is thought that this approach is faster that keeping such objects in separate list
	msg("* merging...", newline = False)
	result = []
	esz = len(entries)
	for i in range(0, esz - 1) :
		e = entries[i]
		if hasattr(e, "__matched") :
			continue
		matching = [e] # List containing matching entries
		for j in range(i + 1, esz) :
			x = entries[j]
			if hasattr(x, "__matched") :
				continue
			# Matching e and x
			if os.path.basename(e.exename) == os.path.basename(x.exename) :
				# The main criteria is the considence of the executables
				try :
					eargs = e.exeargs
				except AttributeError :
					eargs = None
				try :
					xargs = x.exeargs
				except AttributeError :
					xargs = None
				if not eargs and not xargs :
					# Both entries don't have arguments, consider them matching
					matching.append(x)
				elif eargs and xargs :
					# Both entries do have arguments, match them
					if Kw(eargs) == Kw(xargs) :
						# Squeeze the spaces and don't consider the case of the characters
						matching.append(x)
		if len(matching) < 2 :
			# Only one candidate, no choice
			winner = e
		else :
			# Several candidates, must choose the most relevant
			winner = None
			for m in matching :
				if winner :
					# For now the longer name is the better
					if len(m.name) > len(winner.name) :
						winner = m
				else :
					winner = m
		result.append(winner)
		# This is done in the last step to mark e which was put in matching
		for m in matching :
			m.__matched = True
	msg(" %d coincidings detected" % (len(entries) - len(result)))
	return result
